# Reference: http://python-packaging.readthedocs.io/en/latest/dependencies.html

__title__ = 'skater'
__version__ = '1.1.2'
__summary__ = 'Model Interpretation Library'
__uri__ = 'https://github.com/datascienceinc/skater/'
__author__ = 'Aaron Kramer, Pramit Choudhary, DataScience.com Team, and contributors'
__maintainer__ = 'Aaron Kramer, Pramit Choudhary and community'
__email__ = 'dstm@datascience.com'
__license__ = 'MIT'
__docs__ = 'https://datascienceinc.github.io/Skater/'
